#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>

#include "api.h"

void show_usage(char *cmd)
{
    printf("Usage: %s [OPTIONS]\n", cmd);

    printf("\n[OPTIONS]\n");
    printf("  -h, --help              show this usage of api_linux_demo\n");
    printf("  -a, --alarm             test the system alarm\n");
    printf("  -b, --uart              test the system uart\n");
    printf("  -c, --i2c               test the system i2c\n");
    printf("  -d, --spi               test the system spi\n");
    printf("  -e, --can               test the system can\n");
    printf("  -f, --gpio              test the system gpio\n");
    printf("  -g, --gpiointerrupt     test the system gpio interrupt\n");
}

static void gpio_poll_callback(int gpioNumber, int value, const char *reason);

int main(int argc, char *argv[])
{
    int cmd = 0;	// Command code
    int rc = 0;		// Command Return code
    int ret = -1;   // API Return code

    /* Options string */
    const char *short_opt_str = "habcdefg";
    const struct option long_opt_str[] = {
        {"help",	no_argument,	NULL, 'h'},
        {"alarm",	no_argument,	NULL, 'a'},
        {"uart",	no_argument,	NULL, 'b'},
        {"i2c",	    no_argument,	NULL, 'c'},
        {"spi",	    no_argument,	NULL, 'd'},
        {"can",	    no_argument,	NULL, 'e'},
        {"gpio",	no_argument,	NULL, 'f'},
        {"gpiointerrupt",	no_argument,	NULL, 'g'},
        {NULL, 		no_argument,	NULL, '?'}
    };

    /* Show usage if no arguments */
    if (1 == argc) {
        show_usage(argv[0]);
        exit(0);
    }

    /* Parse commandline */
    do {
        rc = getopt_long(argc, argv, short_opt_str, long_opt_str, NULL);
        switch (rc) {
        case -1:
            break;
        case 'h':
            show_usage(argv[0]);
            exit(0);
        case 'a':
        case 'b':
        case 'c':
        case 'd':
        case 'e':
        case 'f':
        case 'g':
            cmd = rc;
            break;
        case '?':
        default:
            printf("Invalid arguments, please refer to the usage with '--help' option.\n");
            exit(-1);
        }
    } while (-1 != rc);

    /* Process command */
    switch (cmd) {

    // 定时开机接口测试
    case 'a':
    {
        // 注意时区，UTC表示0时区，我们处于CST时区比UTC时间早8小时
        // 设置后可以通过shutdown -h now命令关机，来测试定时开机功能
        char date_time[100] = {0};
        char alarm_time[100] = {0};

        rtc_time_get(date_time);
        printf("rtc_time_get date_time=%s\n", date_time);

        rtc_alarm_get(alarm_time);
        printf("rtc_alarm_get1 alarm_time=%s\n", alarm_time);

        // 比如设置为2020-11-23 7:50:00开机
        // int day = 22, hour = 23, minute = 50;

        // 比如设置为2020-11-23 9:50:00开机
        int day = 23, hour = 1, minute = 50;
        ret = rtc_alarm_set(day, hour, minute);
        printf("rtc_alarm_set %d:%d ret=%d\n", hour, minute, ret);

        memset(alarm_time, 0, sizeof(alarm_time));
        rtc_alarm_get(alarm_time);
        printf("rtc_alarm_get2 alarm_time=%s\n", alarm_time);

        // 使能定时开机
        int enabled = 1;
        ret = rtc_alarm_enable(enabled);
        printf("rtc_alarm_enable ret=%d\n", ret);

        break;
    }

    // 串口接口测试
    // 测试RS232自发自收
    case 'b':
    {
        int baudrate = 115200;
        const char *node = "/dev/ttyS0";
        int fd = uart_open_advance(node, baudrate, 0, 8, 1, 'N', 0, 1);
        printf("call uartOpen return fd=%d\n", fd);

        char send_buffer[8] = {0xAA, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0xFF};
        int send_len = uart_send(fd, send_buffer, sizeof(send_buffer));
        printf("call uart_send send_len=%d\n", send_len);

        usleep(500 * 1000);

        // 读到所需要的长度
        int read_len = 8;
        int recv_len = 0, all_read = 0;
        char recv_buf[8] = {0};
        while(1) {
            recv_len = uart_recv(fd, recv_buf, read_len - recv_len);
            all_read = all_read + recv_len;
            if (all_read == read_len) {
                printf("all_read == read_len\n");
                break;
            }
        }

        // 打印接收的数据
        if (all_read > 0) {
            printf("recv_buf=");
            for (int i = 0; i < all_read; i++) {
                printf("0x%.2X ", recv_buf[i]);
            }
            printf("\n");
        }

        uart_close(fd);

        break;
    }

    // i2c接口测试
    case 'c':
    {
        const char *node = "/dev/i2c-4";
        int fd = i2c_open(node);
        if (fd < 0) {
        	printf("Unable to open i2c control file.\n");
        	exit(-1);
        }
        int addr = 0x47;
        unsigned int data = 0x00001600;

        if (i2c_write(fd, addr, 0x00, (unsigned char *)&data, 4) < 0)
            return -1;
        usleep(1000);

        do {
            if (i2c_read(fd, addr, 0x00, (unsigned char *)&data, 4) < 0)
                return -1;
            usleep(1000);
        } while (0x00010001 != data);

        i2c_close(fd);
        break;
    }

    // spi接口测试
    case 'd':
    {
        const char *node = "/dev/spidev4.0";
        printf("call spiOpen node=%s\n", node);
        int fd = spi_open(node);
        if (fd < 0) {
            printf("Unable to open spi control file.\n");
            exit(-1);
        }

        uint8_t tx[] = {
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
            0x40, 0x00, 0x00, 0x00, 0x00, 0x95,
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
            0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
            0xDE, 0xAD, 0xBE, 0xEF, 0xBA, 0xAD,
            0xF0, 0x0D, 0xBE, 0xEF, 0xBA, 0xAD
        };
        #define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
        uint8_t rx[ARRAY_SIZE(tx)] = {0};

        struct spi_ioc_transfer tr = {
            .tx_buf = (unsigned long)tx,
            .rx_buf = (unsigned long)rx,
            .len = ARRAY_SIZE(tx),
            .delay_usecs = 0,
            .speed_hz = 500000,
            .bits_per_word = 8,
        };

        ret = ioctl(fd, SPI_IOC_MESSAGE(1), &tr);
        printf("send spi message ret=%d\n", ret);

        if (ret < 1) {
            printf("can't send spi message\n");
            return -1;
        }

        for (int i = 0; i < ARRAY_SIZE(tx); i++) {
            printf("%.2X ", rx[i]);
            printf("\n");
        }

        spi_close(fd);
        break;
    }

    // can接口测试
    case 'e':
    {
        // 打开设备1
        int fd0 = can_open("can0");
        printf("call canOpen fd=%d\n", fd0);

        // 打开设备2
        int fd1 = can_open("can1");
        printf("call canOpen fd=%d\n", fd1);

        // 设备1发数据(设备1和设备2可以相互收发)
        // 发送的canid
        unsigned long canid = 0x10;
        const char send_buffer[8] = {0x01,0x02,0x03,0x04,0x05,0x26,0x37,0x48};
        int write_len = can_write(fd0, canid, send_buffer, 8);
        printf("call canWrite write_len=%d\n", write_len);

        usleep(1000);

        // 设备2收数据(设备1和设备2可以相互收发)
        unsigned int read_len;
        unsigned long read_canid = 0;
        unsigned char read_buffer[8] = {0};
        can_read(fd1, read_buffer, &read_len, &read_canid);
        printf("call canRead can_frame read_canid=%lu read_len=%u\n", read_canid, read_len);

        // 打印读到的数据
        for (int i = 0; i < read_len; i++) {
            printf("call canRead can_frame read_buffer[%d]=0x%02x\n", i, read_buffer[i]);
        }

        // 关闭设备1,设备2
        can_close(fd0);
        can_close(fd1);
        break;
    }

    // gpio接口测试
    case 'f':
    {
        int gpioNumber = 146;

        // 导出GPIO
        int ret = gpio_export(gpioNumber);
        printf("call gpio_export ret=%d\n", ret);

        // 配置为输出引脚
        gpio_direction(gpioNumber, 1);

        // 读取引脚电平
        int value = gpio_read(gpioNumber);
        printf("call gpio_read x1 value=%d\n", value);

        // 输出高电平
        gpio_write(gpioNumber, 1);

        // 读取引脚电平
        value = gpio_read(gpioNumber);
        printf("call gpio_read x2 value=%d\n", value);

        // 取消导出GPIO
        ret = gpio_unexport(gpioNumber);
        printf("call gpio_unexport ret=%d\n", ret);
        break;
    }

    case 'g':
    {
        int gpioNumber = 146;

        // 导出GPIO
        int ret = gpio_export(gpioNumber);
        printf("call gpio_export ret=%d\n", ret);

        // 配置为输入引脚
        gpio_direction(gpioNumber, 0);

        // 读取引脚电平
        int value = gpio_read(gpioNumber);
        printf("call gpio_read value=%d\n", value);

        /**
         * "none"
         * "rising"
         * "falling"
         * "both"
         */
        // 设置中断触发方式
        ret = gpio_set_edge(gpioNumber, "rising");
        printf("call gpio_set_edge ret=%d\n", ret);

        // 获取设置的中断触发方式
        char edge[8] = {0};
        ret = gpio_get_edge(gpioNumber, edge);
        printf("call gpio_get_edge ret=%d\n", ret);
        if (ret == 0) {
            printf("call gpio_get_edge edge=%s\n", edge);
        }

        int timeout = 200; // 单位为ms
        // 开启中断
        ret = gpio_poll_start(gpioNumber, timeout, gpio_poll_callback);
        printf("call gpio_poll_start ret=%d\n", ret);

        // 模拟外设运行
        sleep(60);

        // 停止中断
        gpio_poll_stop(gpioNumber);

        // 取消导出GPIO
        ret = gpio_unexport(gpioNumber);
        printf("call gpio_unexport ret=%d\n", ret);
        break;
    }

    default:
        break;
    }

    return 0;
}

// GPIO中断数据回调接口
static void gpio_poll_callback(int gpioNumber, int value, const char *reason)
{
    // do work here
    printf("call gpio_poll_callback gpioNumber=%d value=%d reason=%s\n", gpioNumber, value, reason);
}

